// PCMPlayer.cpp
//

#include "PCMPlayer.h"
#include <process.h>
#include <stdio.h>

namespace {
	unsigned int playTime_g = 1;	// 1 sec.
}

namespace Dix {
	PCMPlayer::PCMPlayer() :
		isReady_		( false ),
		threadHandle_	( 0 ),
		isTerminate_	( false ),
		isLoop_			( true ),
		state_			( STATE_NONE )
	{
		clear();
	}

	PCMPlayer::PCMPlayer( Com_ptr< IDirectSound8 > cpDS8 ) :
		cpDS8_			( cpDS8 ),
		isReady_		( false ),
		threadHandle_	( 0 ),
		isTerminate_	( false ),
		isLoop_			( true ),
		state_			( STATE_NONE )
	{
		clear();
	}

	PCMPlayer::PCMPlayer( Com_ptr< IDirectSound8 > cpDS8, sp< PCMDecoder > spDecoder ) :
		cpDS8_			( cpDS8 ),
		isReady_		( false ),
		threadHandle_	( 0 ),
		isTerminate_	( false ),
		isLoop_			( true ),
		state_			( STATE_NONE )
	{
		clear();
		setDecoder( spDecoder );
	}

	PCMPlayer::~PCMPlayer() {
		terminateThread();
	}

	//! NA
	void PCMPlayer::clear() {
		terminateThread();
		memset( &DSBufferDesc_, 0, sizeof( DSBufferDesc_ ) );
		memset( &waveFormat_, 0, sizeof( waveFormat_ ) );
		cpDSBuffer_ = 0;
		isReady_ = false;
		state_ = STATE_NONE;
	}

	//! Đ̃Xbh~
	void PCMPlayer::terminateThread() {
		isTerminate_ = true;
		if ( threadHandle_ != 0 ) {
			bool end = false;
			while( !end ) {
				DWORD flag = WaitForSingleObject( (HANDLE)(__int64)threadHandle_, 100 );
				switch( flag ) {
				case WAIT_OBJECT_0:
					// XbhI
					end = true;
					break;
				case WAIT_TIMEOUT:
					// ܂IĂȂ̂őҋ@
					break;
				case WAIT_FAILED:
					// sĂ悤ł
					end = true;
					break;
				}
			}
		}
		isTerminate_ = false;
		threadHandle_ = 0;
	}

	//! foCXݒ
	void PCMPlayer::setDevice( Com_ptr< IDirectSound8 > cpDS8 ) {
		cpDS8_ = cpDS8;
	}

	//! PCMfR[_ݒ
	bool PCMPlayer::setDecoder( sp< PCMDecoder > pcmDecoder ) {
		if ( cpDS8_.GetPtr() == 0 || pcmDecoder.GetPtr() == 0 || pcmDecoder->isReady() == false ) {
			isReady_ = false;
			return false;
		}

		state_ = STATE_STOP;

		if ( !pcmDecoder->getWaveFormatEx( waveFormat_ ) ) {
			return false;
		}

		DSBufferDesc_.dwSize = sizeof( DSBUFFERDESC );
		DSBufferDesc_.dwFlags = DSBCAPS_CTRLPAN | DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLFREQUENCY | DSBCAPS_GLOBALFOCUS | DSBCAPS_CTRLPOSITIONNOTIFY;
		DSBufferDesc_.dwBufferBytes = waveFormat_.nAvgBytesPerSec * playTime_g;
		DSBufferDesc_.dwReserved = 0;
		DSBufferDesc_.lpwfxFormat = &waveFormat_;
		DSBufferDesc_.guid3DAlgorithm = GUID_NULL;

		// N[ۑ
		spPCMDecoder_ = pcmDecoder->createClone();

		// ZJ_obt@܂ꍇ͍쐬
		if ( cpDSBuffer_.GetPtr() == 0 ) {
			IDirectSoundBuffer*	 ptmpBuf = 0;
			if ( SUCCEEDED( cpDS8_->CreateSoundBuffer( &DSBufferDesc_, &ptmpBuf, NULL ) ) ) {
				ptmpBuf->QueryInterface( IID_IDirectSoundBuffer8 , (void**)cpDSBuffer_.ToCreator() );
			}
			else {
				clear();
				return false;
			}
			ptmpBuf->Release();
		}

		// obt@
		if ( initializeBuffer() == false ) {
			return false;
		}

		// obt@Rs[Xbh
		if ( threadHandle_ == 0 ) {
			threadHandle_ = (unsigned int)_beginthread( PCMPlayer::streamThread, 0, (void*)this );
		}

		isReady_ = true;

		return true;
	}

	//! obt@
	bool PCMPlayer::initializeBuffer() {
		if ( spPCMDecoder_.GetPtr() == 0 ) {
			return false;
		}

		spPCMDecoder_->setHead();	// o
		cpDSBuffer_->SetCurrentPosition( 0 );

		// obt@bNďf[^
		void* AP1 = 0, *AP2 = 0;
		DWORD AB1 = 0, AB2  = 0;
		if ( SUCCEEDED( cpDSBuffer_->Lock( 0, 0, &AP1, &AB1, &AP2, &AB2, DSBLOCK_ENTIREBUFFER ) ) ) {
			spPCMDecoder_->getSegment( (char*)AP1, AB1, 0, 0 );
			cpDSBuffer_->Unlock( AP1, AB1, AP2, AB2 );
		}
		else {
			clear();
			return false;
		}

		return true;
	}

	//! Xg[ĐXbh
	void PCMPlayer::streamThread( void* playerPtr ) {
		PCMPlayer* player = (PCMPlayer*)playerPtr;
		unsigned int size = player->DSBufferDesc_.dwBufferBytes / 2;
		unsigned int flag = 0;
		DWORD point = 0;
		void* AP1 = 0, *AP2 = 0;
		DWORD AB1 = 0, AB2  = 0;

		DWORD finishPos = 0;
		DWORD prePlayPos = 0;

		unsigned int writeSize = 0;
		bool isEnd = false;
		bool waitFinish = false;

		while( player->isTerminate_ == false ) {
			switch ( player->getState() ) {
			case STATE_PLAY:	// Đ
				// Xg[Đ
				// ݈ʒu`FbN
				player->cpDSBuffer_->GetCurrentPosition( &point, 0 );
				if ( flag == 0 && point >= size ) {
					// Oɏ
					if ( SUCCEEDED( player->cpDSBuffer_->Lock( 0, size, &AP1, &AB1, &AP2, &AB2, 0 ) ) ) {
						player->spPCMDecoder_->getSegment( (char*)AP1, AB1, &writeSize, &isEnd );
						player->cpDSBuffer_->Unlock( AP1, AB1, AP2, AB2 );
						flag = 1;
					}

					// ŏȈ݂ꍇ͏Iʒu
					if ( isEnd == true && waitFinish == false ) {
						finishPos = writeSize;
						player->cpDSBuffer_->GetCurrentPosition( &prePlayPos, 0 );
						waitFinish = true;
					}
				}
				else if ( flag == 1 && point < size ) {
					// 㔼ɏ
					if ( SUCCEEDED( player->cpDSBuffer_->Lock( size, size * 2, &AP1, &AB1, &AP2, &AB2, 0 ) ) ) {
						player->spPCMDecoder_->getSegment( (char*)AP1, AB1, &writeSize, &isEnd );
						player->cpDSBuffer_->Unlock( AP1, AB1, AP2, AB2 );
						flag = 0;
					}

					// ŏȈ݂ꍇNorifyݒ
					if ( isEnd == true && waitFinish == false ) {
						finishPos = size + writeSize;
						player->cpDSBuffer_->GetCurrentPosition( &prePlayPos, 0 );
						waitFinish = true;
					}
				}
				break;

			case STATE_STOP:
				flag = 0;	// ~߂ƑO݂n܂邽
				isEnd = false;
				finishPos = 0;
				prePlayPos = 0;
				break;

			case STATE_PAUSE:
				break;

			default:
				break;
			}

			// Iʒu`FbN
			if ( isEnd == true ) {
				DWORD curPlayPos;
				player->cpDSBuffer_->GetCurrentPosition( &curPlayPos, 0 );
				if ( curPlayPos < prePlayPos ) {
					// obt@[vu
					if ( prePlayPos <= finishPos || finishPos <= curPlayPos ) {
						// I錾
						player->stop();
					}
				} else {
					if ( prePlayPos <= finishPos && finishPos <= curPlayPos ) {
						// I錾
						player->stop();
					}
				}
				prePlayPos = curPlayPos;
			}

			Sleep( 100 );
		}
	}

	//! Đ
	bool PCMPlayer::play( bool isLoop ) {
		if ( isReady() == false ) {
			return false;
		}
		isLoop_ = isLoop;
		spPCMDecoder_->setLoop( isLoop );
		cpDSBuffer_->Play( 0, 0, DSBPLAY_LOOPING );
		state_ = STATE_PLAY;
		return true;
	}

	//! ꎞ~
	void PCMPlayer::pause() {
		if ( state_ == STATE_PLAY ) {
			// Ă~߂
			cpDSBuffer_->Stop();
			state_ = STATE_PAUSE;
		}
		else {
			// ~܂ĂĐ
			play( isLoop_ );
		}
	}

	//! ~
	void PCMPlayer::stop() {
		if ( isReady() == false ) {
			return;
		}
		state_ = STATE_STOP;
		cpDSBuffer_->Stop();

		// obt@̓o
		initializeBuffer();
	}

	//! ʂς
	void PCMPlayer::setVolume( int volume ) {
		if ( isReady() == true ) {
			cpDSBuffer_->SetVolume( volume );
		}
	}

	//! p̈ʒuς
	void PCMPlayer::setPan( int pan ) {
		if ( isReady() == true ) {
			cpDSBuffer_->SetPan( pan );
		}
	}

	//! łH
	bool PCMPlayer::isReady() {
		return isReady_;
	}

	//! ĐH
	bool PCMPlayer::isPlaying() {
		return ( state_ == STATE_PLAY );
	}

	//! Ԃ擾
	PCMPlayer::STATE PCMPlayer::getState() {
		return state_;
	}

}